/*
 * vifd/vif_notify.c
 * $Rev $
 *
 * Copyright (C) 2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/**
 * @file vif_notify.c
 * @brief Functions for handling notification message queue access
 */

/* INCLUDES *******************************************************************/

#include <netinet/in.h>                 // struct in_addr
#include <linux/if.h>                   // IFNAMSIZ
#include <sys/ipc.h>                    // IPC_CREATE
#include <sys/msg.h>                    // msgget(), msgrecv(), msgsnd(), ...
#include <errno.h>                      // errno
#include <stdio.h>                      // printf()
#include <string.h>                     // memset(), strerror()
#include "vif_notify.h"


/* MESSAGE QUEUE TYPES ********************************************************/

#define VIF_NOTIFY_KEY      1337    /**< Name of notification message queue */

/** Message format to indicate a ioctl() */
typedef struct {
    long mtype;                     /**< Message type */
    int ioctl_type;                 /**< ioctl() type */
    char dev_name[IFNAMSIZ + 1];    /**< Virtual interface device name */
} ioctl_msg;

#define IOCTL_MSG_SIZE      sizeof(ioctl_msg)
#define VIF_SIOCSIW         0x10    /**< Message type: ioctl() set wireless parameter */


/* OUTPUT *********************************************************************/

int
ignore_debug(const char* text, ...)
{
    return 0;
}

/** Debug output function (ignored by default);
 *  can be changed with vif_notify_set_debug_function() */
int (*vif_notify_debug)(const char *format, ...) = &ignore_debug;

/** Error output function (default to stdout);
 *  can be changed with vif_notify_set_error_function() */
int (*vif_notify_error)(const char *format, ...) = &printf;

void
vif_notify_set_debug_function(int (*debug)(const char *format, ...))
{
    vif_notify_debug = debug;
}

void
vif_notify_set_error_function(int (*error)(const char *format, ...))
{
    vif_notify_error = error;
}


/* PRIVATE FUNCTIONS (DON'T CALL FROM OUTSIDE) ********************************/

/**
 * Attach to the notification message queue
 * @param flags Message queue flags (see: man msgget)
 * @return Message queue ID
 */
int
vif_notify_attach(int flags)
{
    int msgq_id = msgget(VIF_NOTIFY_KEY, flags);
    if (msgq_id < 0)
    {
        vif_notify_error("vif_notify_attach(): msgget (%s)\n", strerror(errno));
        return -1;
    }

    //>>>>>>>>>> DEBUG <<<<<<<<<<<<<<
    vif_notify_debug("vif_notify_attach(): msgq_id = %i\n", msgq_id);
    return msgq_id;
}


/* PUBLIC FUNCTIONS ***********************************************************/

int
vif_notify_init()
{
    int create_flags = IPC_CREAT | 0664;                  // read/write queue

    return vif_notify_attach(create_flags);
}

int
vif_notify_remove(int msgq_id)
{
    return msgctl(msgq_id, IPC_RMID, NULL);;
}

int
vif_notify_ioctl(int ioctl_type, const char *dev_name)
{
    ioctl_msg msg;
    int write_flags = 0600;

    int msgq_id = vif_notify_attach(write_flags);
    if (msgq_id < 0)
        return -1;

    // prepare message (fill type and vif name)
    memset(&msg, 0, IOCTL_MSG_SIZE);
    msg.mtype = VIF_SIOCSIW;
    msg.ioctl_type = ioctl_type;
    strncpy(msg.dev_name, dev_name, IFNAMSIZ);
    msg.dev_name[IFNAMSIZ] = '\0';          // make sure we have proper string

    if (msgsnd(msgq_id, &msg, IOCTL_MSG_SIZE, 0) < 0)
    {
        vif_notify_error("vif_notify_ioctl(): msgsnd (%s)\n", strerror(errno));
        return -1;
    }

    return 0;
}

void
vif_notify_poll(void *msg_handler)
{
    ioctl_msg in_buffer;
    int read_flags = 0400;
    int msgq_id = -1;

    if (msg_handler == NULL)
    {
        vif_notify_error("vif_notify_init(): msg_handler == NULL\n");
        return;
    }
    if ((msgq_id = vif_notify_attach(read_flags)) < 0)
        return;

    memset(&in_buffer, 0, sizeof(ioctl_msg));
    while (1)
    {
        if (msgrcv(msgq_id, &in_buffer, IOCTL_MSG_SIZE, VIF_SIOCSIW, 0) < 0)
        {
            vif_notify_error("vif_notify_poll(): msgrcv (%s)\n", strerror(errno));
            return;
        }
        // now correctly cast the the function pointer and execute
        ((void (*)(int ioctl_type, const char *dev_name)) msg_handler)
            (in_buffer.ioctl_type, in_buffer.dev_name);
    }
    return; // this should never be reached!
}
